@@ -29,21 +29,7 @@
 
 #include "mdadm.h"
 #include "md_p.h"
-
-int open_mddev(char *dev)
-{
-	int mdfd = open(dev, O_RDWR, 0);
-	if (mdfd < 0)
-		fprintf(stderr, Name ": error opening %s: %s\n",
-			dev, strerror(errno));
-	else if (md_get_version(mdfd) <= 0) {
-		fprintf(stderr, Name ": %s does not appear to be an md device\n",
-			dev);
-		close(mdfd);
-		mdfd = -1;
-	}
-	return mdfd;
-}
+#include <ctype.h>
 
 
 
@@ -57,9 +43,9 @@ int main(int argc, char *argv[])
 	int rv;
 
 	int chunk = 0;
-	int size = 0;
-	int level = -10;
-	int layout = -1;
+	int size = -1;
+	int level = UnSet;
+	int layout = UnSet;
 	int raiddisks = 0;
 	int sparedisks = 0;
 	struct mddev_ident_s ident;
@@ -78,17 +64,25 @@ int main(int argc, char *argv[])
 	int verbose = 0;
 	int brief = 0;
 	int force = 0;
+	int test = 0;
+	int assume_clean = 0;
+	int autof = 0; /* -1 for non-partitions, 1 or more to create partitions */
 
 	char *mailaddr = NULL;
 	char *program = NULL;
 	int delay = 0;
+	int daemonise = 0;
+	char *pidfile = NULL;
+	int oneshot = 0;
+
+	int copies;
 
 	int mdfd = -1;
 
 	ident.uuid_set=0;
-	ident.level = -10;
-	ident.raid_disks = -1;
-	ident.super_minor= -1;
+	ident.level = UnSet;
+	ident.raid_disks = UnSet;
+	ident.super_minor= UnSet;
 	ident.devices=0;
 
 	while ((option_index = -1) ,
@@ -100,14 +94,19 @@ int main(int argc, char *argv[])
 		switch(opt) {
 		case 'h':
 			help_text = Help;
-			switch (mode) {
-			case ASSEMBLE : help_text = Help_assemble; break;
-			case BUILD    : help_text = Help_build; break;
-			case CREATE   : help_text = Help_create; break;
-			case MANAGE   : help_text = Help_manage; break;
-			case MISC     : help_text = Help_misc; break;
-			case MONITOR  : help_text = Help_monitor; break;
-			}
+			if (option_index > 0 && 
+			    strcmp(long_options[option_index].name, "help-options")==0)
+				help_text = OptionHelp;
+			else
+				switch (mode) {
+				case ASSEMBLE : help_text = Help_assemble; break;
+				case BUILD    : help_text = Help_build; break;
+				case CREATE   : help_text = Help_create; break;
+				case MANAGE   : help_text = Help_manage; break;
+				case MISC     : help_text = Help_misc; break;
+				case MONITOR  : help_text = Help_monitor; break;
+				case GROW     : help_text = Help_grow; break;
+				}
 			fputs(help_text,stderr);
 			exit(0);
 
@@ -143,6 +142,7 @@ int main(int argc, char *argv[])
 		case 'B': newmode = BUILD; break;
 		case 'C': newmode = CREATE; break;
 		case 'F': newmode = MONITOR;break;
+		case 'G': newmode = GROW; break;
 
 		case '#':
 		case 'D':
@@ -193,13 +193,18 @@ int main(int argc, char *argv[])
 		case 'B':
 		case 'C':
 		case 'F':
+		case 'G':
 			continue;
 		}
 		if (opt == 1) {
 		        /* an undecorated option - must be a device name.
 			 */
 			if (devs_found > 0 && mode == '@' && !devmode) {
-				fprintf(stderr, Name ": Must give on of -a/-r/-f for subsequent devices at %s\n", optarg);
+				fprintf(stderr, Name ": Must give one of -a/-r/-f for subsequent devices at %s\n", optarg);
+				exit(2);
+			}
+			if (devs_found > 0 && mode == 'G' && !devmode) {
+				fprintf(stderr, Name ": Must give one of -a for devices do add: %s\n", optarg);
 				exit(2);
 			}
 			dv = malloc(sizeof(*dv));
@@ -236,34 +241,40 @@ int main(int argc, char *argv[])
 			}
 			continue;
 
+		case O(GROW,'z'):
 		case O(CREATE,'z'): /* size */
-			if (size) {
+			if (size >= 0) {
 				fprintf(stderr, Name ": size may only be specified once. "
 					"Second value is %s.\n", optarg);
 				exit(2);
 			}
-			size = strtol(optarg, &c, 10);
-			if (!optarg[0] || *c || size < 4) {
-				fprintf(stderr, Name ": invalid size: %s\n",
-					optarg);
-				exit(2);
+			if (strcmp(optarg, "max")==0)
+				size = 0;
+			else {
+				size = strtol(optarg, &c, 10);
+				if (!optarg[0] || *c || size < 4) {
+					fprintf(stderr, Name ": invalid size: %s\n",
+						optarg);
+					exit(2);
+				}
 			}
 			continue;
 
+		case O(GROW,'l'): /* hack - needed to understand layout */
 		case O(CREATE,'l'):
 		case O(BUILD,'l'): /* set raid level*/
-			if (level != -10) {
+			if (level != UnSet) {
 				fprintf(stderr, Name ": raid level may only be set once.  "
 					"Second value is %s.\n", optarg);
 				exit(2);
 			}
 			level = map_name(pers, optarg);
-			if (level == -10) {
+			if (level == UnSet) {
 				fprintf(stderr, Name ": invalid raid level: %s\n",
 					optarg);
 				exit(2);
 			}
-			if (level != 0 && level != -1 && mode == BUILD) {
+			if (level != 0 && level != -1 && level != 1 && level != -4 && level != -5 && mode == BUILD) {
 				fprintf(stderr, Name ": Raid level %s not permitted with --build.\n",
 					optarg);
 				exit(2);
@@ -277,7 +288,9 @@ int main(int argc, char *argv[])
 			continue;
 
 		case O(CREATE,'p'): /* raid5 layout */
-			if (layout >= 0) {
+		case O(BUILD,'p'): /* faulty layout */
+		case O(GROW, 'p'): /* faulty reconfig */
+			if (layout != UnSet) {
 				fprintf(stderr,Name ": layout may only be sent once.  "
 					"Second value was %s\n", optarg);
 				exit(2);
@@ -287,21 +300,60 @@ int main(int argc, char *argv[])
 				fprintf(stderr, Name ": layout not meaningful for %s arrays.\n",
 					map_num(pers, level));
 				exit(2);
-			case -10:
+			case UnSet:
 				fprintf(stderr, Name ": raid level must be given before layout.\n");
 				exit(2);
 
 			case 5:
+			case 6:
 				layout = map_name(r5layout, optarg);
-				if (layout==-10) {
+				if (layout==UnSet) {
 					fprintf(stderr, Name ": layout %s not understood for raid5.\n",
 						optarg);
 					exit(2);
 				}
 				break;
+
+			case 10:
+				/* 'f' or 'n' followed by a number <= raid_disks */
+				if ((optarg[0] !=  'n' && optarg[0] != 'f') ||
+				    (copies = strtoul(optarg+1, &cp, 10)) < 1 ||
+				    copies > 200 ||
+				    *cp) {
+					fprintf(stderr, Name ": layout for raid10 must be 'nNN' or 'fNN' where NN is a number, not %s\n", optarg);
+					exit(2);
+				}
+				if (optarg[0] == 'n')
+					layout = 256 + copies;
+				else
+					layout = 1 + (copies<<8);
+				break;
+			case -5: /* Faulty
+				  * modeNNN
+				  */
+				    
+			{
+				int ln = strcspn(optarg, "0123456789");
+				char *m = strdup(optarg);
+				int mode;
+				m[ln] = 0;
+				mode = map_name(faultylayout, m);
+				if (mode == UnSet) {
+					fprintf(stderr, Name ": layout %s not understood for faulty.\n",
+						optarg);
+					exit(2);
+				}
+				layout = mode | (atoi(optarg+ln)<< ModeShift);
+			}
 			}
 			continue;
 
+		case O(CREATE,3):
+		case O(BUILD,3): /* assume clean */
+			assume_clean = 1;
+			continue;
+
+		case O(GROW,'n'):
 		case O(CREATE,'n'):
 		case O(BUILD,'n'): /* number of raid disks */
 			if (raiddisks) {
@@ -315,6 +367,12 @@ int main(int argc, char *argv[])
 					optarg);
 				exit(2);
 			}
+			if (raiddisks == 1 &&  !force && level != -5) {
+				fprintf(stderr, Name ": '1' is an unusual number of drives for an array, so it is probably\n"
+					"     a mistake.  If you really mean it you will need to specify --force before\n"
+					"     setting the number of drives.\n");
+				exit(2);
+			}
 			ident.raid_disks = raiddisks;
 			continue;
 
@@ -324,7 +382,7 @@ int main(int argc, char *argv[])
 					sparedisks, optarg);
 				exit(2);
 			}
-			if (level > -10 && level <= 0 && level >= -1) {
+			if (level != UnSet && level <= 0 && level >= -1) {
 				fprintf(stderr, Name ": spare-devices setting is incompatible with raid level %d\n",
 					level);
 				exit(2);
@@ -336,6 +394,44 @@ int main(int argc, char *argv[])
 				exit(2);
 			}
 			continue;
+
+		case O(CREATE,'a'):
+		case O(BUILD,'a'):
+		case O(ASSEMBLE,'a'): /* auto-creation of device node */
+			if (optarg == NULL)
+				autof = -1;
+			else if (strcasecmp(optarg,"no")==0)
+				autof = 0;
+			else if (strcasecmp(optarg,"yes")==0 || strcasecmp(optarg,"md")==0)
+				autof = -1;
+			else {
+				/* There might be digits, and maybe a hypen, at the end */
+				char *e = optarg + strlen(optarg);
+				int num = 4;
+				int len;
+				while (e > optarg && isdigit(e[-1]))
+					e--;
+				if (*e) {
+					num = atoi(e);
+					if (num <= 0) num = 1;
+				}
+				if (e > optarg && e[-1] == '-')
+					e--;
+				len = e - optarg;
+				if ((len == 3 && strncasecmp(optarg,"mdp",3)==0) ||
+				    (len == 1 && strncasecmp(optarg,"p",1)==0) ||
+				    (len >= 4 && strncasecmp(optarg,"part",4)==0))
+					autof = num;
+				else {
+					fprintf(stderr, Name ": --auto flag arg of \"%s\" unrecognised: use no,yes,md,mdp,part\n"
+						"        optionally followed by a number.\n",
+						optarg);
+					exit(2);
+				}
+			}
+			continue;
+
+		case O(BUILD,'f'): /* force honouring '-n 1' */
 		case O(CREATE,'f'): /* force honouring of device list */
 		case O(ASSEMBLE,'f'): /* force assembly */
 		case O(MISC,'f'): /* force zero */
@@ -358,15 +454,19 @@ int main(int argc, char *argv[])
 			continue;
 
 		case O(ASSEMBLE,'m'): /* super-minor for array */
-			if (ident.super_minor >= 0) {
+			if (ident.super_minor != UnSet) {
 				fprintf(stderr, Name ": super-minor cannot be set twice.  "
 					"Second value: %s.\n", optarg);
 				exit(2);
 			}
-			ident.super_minor = strtoul(optarg, &cp, 10);
-			if (!optarg[0] || *cp) {
-				fprintf(stderr, Name ": Bad super-minor number: %s.\n", optarg);
-				exit(2);
+			if (strcmp(optarg, "dev")==0)
+				ident.super_minor = -2;
+			else {
+				ident.super_minor = strtoul(optarg, &cp, 10);
+				if (!optarg[0] || *cp) {
+					fprintf(stderr, Name ": Bad super-minor number: %s.\n", optarg);
+					exit(2);
+				}
 			}
 			continue;
 
@@ -377,10 +477,15 @@ int main(int argc, char *argv[])
 				exit(2);
 			}
 			update = optarg;
-			if (strcmp(update, "sparc2.2")==0) continue;
+			if (strcmp(update, "sparc2.2")==0) 
+				continue;
 			if (strcmp(update, "super-minor") == 0)
 				continue;
-			fprintf(stderr, Name ": '--update %s' invalid.  Only 'sparc2.2' or 'super-minor' supported\n",update);
+			if (strcmp(update, "summaries")==0)
+				continue;
+			if (strcmp(update, "resync")==0)
+				continue;
+			fprintf(stderr, Name ": '--update %s' invalid.  Only 'sparc2.2', 'super-minor', 'resync' or 'summaries' supported\n",update);
 			exit(2);
 
 		case O(ASSEMBLE,'c'): /* config file */
@@ -429,15 +534,28 @@ int main(int argc, char *argv[])
 				}
 			}
 			continue;
-			
+		case O(MONITOR,'f'): /* daemonise */
+			daemonise = 1;
+			continue;
+		case O(MONITOR,'i'): /* pid */
+			if (pidfile)
+				fprintf(stderr, Name ": only specify one pid file. %s ignored.\n",
+					optarg);
+			else
+				pidfile = optarg;
+			continue;
+		case O(MONITOR,'1'): /* oneshot */
+			oneshot = 1;
+			continue;
+		case O(MONITOR,'t'): /* test */
+			test = 1;
+			continue;
 
 			/* now the general management options.  Some are applicable
 			 * to other modes. None have arguments.
 			 */
-		case O(MANAGE,'a'):
-		case O(CREATE,'a'):
-		case O(BUILD,'a'):
-		case O(ASSEMBLE,'a'): /* add a drive */
+		case O(GROW,'a'):
+		case O(MANAGE,'a'): /* add a drive */
 			devmode = 'a';
 			continue;
 		case O(MANAGE,'r'): /* remove a drive */
@@ -495,6 +613,9 @@ int main(int argc, char *argv[])
 			}
 			devmode = opt;
 			continue;
+		case O(MISC,'t'):
+			test = 1;
+			continue;
 
 		case O(MISC, 22):
 			if (devmode != 'E') {
@@ -527,14 +648,24 @@ int main(int argc, char *argv[])
 	 * we check that here and open it.
 	 */
 
-	if (mode==MANAGE || mode == BUILD || mode == CREATE || (mode == ASSEMBLE && ! scan)) {
+	if (mode==MANAGE || mode == BUILD || mode == CREATE || mode == GROW ||
+	    (mode == ASSEMBLE && ! scan)) {
 		if (devs_found < 1) {
 			fprintf(stderr, Name ": an md device must be given in this mode\n");
 			exit(2);
 		}
-		mdfd = open_mddev(devlist->devname);
+		if ((int)ident.super_minor == -2 && autof) {
+			fprintf(stderr, Name ": --super-minor=dev is incompatible with --auto\n");	
+			exit(2);
+		}
+		mdfd = open_mddev(devlist->devname, autof);
 		if (mdfd < 0)
 			exit(1);
+		if ((int)ident.super_minor == -2) {
+			struct stat stb;
+			fstat(mdfd, &stb);
+			ident.super_minor = MINOR(stb.st_rdev);
+		}
 	}
 
 	rv = 0;
@@ -552,7 +683,26 @@ int main(int argc, char *argv[])
 			rv = Manage_runstop(devlist->devname, mdfd, runstop);
 		break;
 	case ASSEMBLE:
-		if (!scan)
+		if (devs_found == 1 && ident.uuid_set == 0 &&
+		    ident.super_minor == UnSet && !scan ) {
+			/* Only a device has been given, so get details from config file */
+			mddev_ident_t array_ident = conf_get_ident(configfile, devlist->devname);
+			if (array_ident == NULL) {
+				fprintf(stderr, Name ": %s not identified in config file.\n",
+					devlist->devname);
+				rv |= 1;
+			} else {
+				mdfd = open_mddev(devlist->devname, array_ident->autof);
+				if (mdfd < 0)
+					rv |= 1;
+				else {
+					rv |= Assemble(devlist->devname, mdfd, array_ident, configfile,
+						       NULL,
+						       readonly, runstop, update, verbose, force);
+					close(mdfd);
+				}
+			}
+		} else if (!scan)
 			rv = Assemble(devlist->devname, mdfd, &ident, configfile,
 				      devlist->next,
 				      readonly, runstop, update, verbose, force);
@@ -563,20 +713,21 @@ int main(int argc, char *argv[])
 			}
 			for (dv = devlist ; dv ; dv=dv->next) {
 				mddev_ident_t array_ident = conf_get_ident(configfile, dv->devname);
-				mdfd = open_mddev(dv->devname);
-				if (mdfd < 0) {
-					rv |= 1;
-					continue;
-				}
 				if (array_ident == NULL) {
 					fprintf(stderr, Name ": %s not identified in config file.\n",
 						dv->devname);
 					rv |= 1;
 					continue;
 				}
+				mdfd = open_mddev(dv->devname, array_ident->autof);
+				if (mdfd < 0) {
+					rv |= 1;
+					continue;
+				}
 				rv |= Assemble(dv->devname, mdfd, array_ident, configfile,
 					       NULL,
 					       readonly, runstop, update, verbose, force);
+				close(mdfd);
 			}
 		} else {
 			mddev_ident_t array_list =  conf_get_ident(configfile, NULL);
@@ -586,7 +737,7 @@ int main(int argc, char *argv[])
 			} else
 				for (; array_list; array_list = array_list->next) {
 					mdu_array_info_t array;
-					mdfd = open_mddev(array_list->devname);
+					mdfd = open_mddev(array_list->devname, array_list->autof);
 					if (mdfd < 0) {
 						rv |= 1;
 						continue;
@@ -602,10 +753,10 @@ int main(int argc, char *argv[])
 		}
 		break;
 	case BUILD:
-		rv = Build(devlist->devname, mdfd, chunk, level, raiddisks, devlist->next);
+		rv = Build(devlist->devname, mdfd, chunk, level, layout, raiddisks, devlist->next, assume_clean);
 		break;
 	case CREATE:
-		rv = Create(devlist->devname, mdfd, chunk, level, layout, size,
+		rv = Create(devlist->devname, mdfd, chunk, level, layout, size<0 ? 0 : size,
 			    raiddisks, sparedisks,
 			    devs_found-1, devlist->next, runstop, verbose, force);
 		break;
@@ -627,7 +778,7 @@ int main(int argc, char *argv[])
 			if (devlist == NULL) {
 				if ((devmode == 'S' ||devmode=='D') && scan) {
 					/* apply to all devices in /proc/mdstat */
-					struct mdstat_ent *ms = mdstat_read();
+					struct mdstat_ent *ms = mdstat_read(0);
 					struct mdstat_ent *e;
 					for (e=ms ; e ; e=e->next) {
 						char *name = get_md_name(e->devnum);
@@ -638,9 +789,9 @@ int main(int argc, char *argv[])
 							continue;
 						}
 						if (devmode == 'D')
-							rv |= Detail(name, !verbose);
+							rv |= Detail(name, !verbose, test);
 						else if (devmode=='S') {
-							mdfd = open_mddev(name);
+							mdfd = open_mddev(name, 0);
 							if (mdfd >= 0)
 								rv |= Manage_runstop(name, mdfd, -1);
 						}
@@ -654,13 +805,13 @@ int main(int argc, char *argv[])
 			for (dv=devlist ; dv; dv=dv->next) {
 				switch(dv->disposition) {
 				case 'D':
-					rv |= Detail(dv->devname, brief); continue;
+					rv |= Detail(dv->devname, brief, test); continue;
 				case 'K': /* Zero superblock */
 					rv |= Kill(dv->devname, force); continue;
 				case 'Q':
 					rv |= Query(dv->devname); continue;
 				}
-				mdfd = open_mddev(dv->devname);
+				mdfd = open_mddev(dv->devname, 0);
 				if (mdfd>=0)
 					switch(dv->disposition) {
 					case 'R':
@@ -676,14 +827,44 @@ int main(int argc, char *argv[])
 		}
 		break;
 	case MONITOR:
-/*
 		if (!devlist && !scan) {
 			fprintf(stderr, Name ": Cannot monitor: need --scan or at least one device\n");
 			rv = 1;
 			break;
 		}
-*/		rv= Monitor(devlist, mailaddr, program,
-			    delay?delay:60, scan, configfile);
+		if (pidfile && !daemonise) {
+			fprintf(stderr, Name ": Cannot write a pid file when not in daemon mode\n");
+			rv = 1;
+			break;
+		}
+		rv= Monitor(devlist, mailaddr, program,
+			    delay?delay:60, daemonise, scan, oneshot, configfile, test, pidfile);
+		break;
+
+	case GROW:
+		if (devs_found > 1) {
+			
+			/* must be '-a'. */
+			if (size >= 0 || raiddisks) {
+				fprintf(stderr, Name ": --size, --raiddisks, and --add are exclusing in --grow mode\n");
+				rv = 1;
+				break;
+			}
+			for (dv=devlist->next; dv ; dv=dv->next) {
+				rv = Grow_Add_device(devlist->devname, mdfd, dv->devname);
+				if (rv)
+					break;
+			}
+		} else if ((size >= 0) + (raiddisks != 0) +  (layout != UnSet) > 1) {
+			fprintf(stderr, Name ": can change at most one of size, raiddisks, and layout\n");
+			rv = 1;
+			break;
+		} else if (layout != UnSet)
+			rv = Manage_reconfig(devlist->devname, mdfd, layout);
+		else if (size >= 0 || raiddisks)
+			rv = Manage_resize(devlist->devname, mdfd, size, raiddisks);
+		else 
+			fprintf(stderr, Name ": no changes to --grow\n");
 		break;
 	}
 	exit(rv);
